import contextlib
from collections import OrderedDict
from typing import Any, Optional

from sag_py_cache_decorator.cache.key import KEY


class LRU(OrderedDict):  # type: ignore
    """A Least recently used cache.
    The longest unused cached parameter result is deleted after the max size is reached.
    """

    def __init__(self, maxsize: Optional[int], *args: Any, **kwargs: Any) -> None:
        self.maxsize = maxsize
        super().__init__(*args, **kwargs)

    def remove_by_key(self, key: KEY) -> None:
        with contextlib.suppress(KeyError):
            del self[key]

    def __getitem__(self, key: KEY) -> Any:
        value = super().__getitem__(key)
        self.move_to_end(key)
        return value

    def __setitem__(self, key: KEY, value: Any) -> None:
        super().__setitem__(key, value)
        self._remove_oldest_if_max_reached()

    def _remove_oldest_if_max_reached(self) -> None:
        if self.maxsize and len(self) > self.maxsize:
            oldest = next(iter(self))
            self.remove_by_key(oldest)



import inspect
import logging
from typing import Any, Callable, Optional, TypeVar, cast

from sag_py_cache_decorator.cache.key import KEY
from sag_py_cache_decorator.cache.lru import LRU

logger = logging.getLogger(__name__)

# With python 3.10 param spec can be used instead - as described here:
# https://stackoverflow.com/questions/66408662/type-annotations-for-decorators
F = TypeVar("F", bound=Callable[..., Any])


def lru_cache(maxsize: Optional[int] = 128) -> Callable[[F], F]:
    cache = LRU(maxsize=maxsize)

    def decorator(func: F) -> F:
        if inspect.iscoroutinefunction(func):

            async def wrapper_async(
                *args: Any,
                lru_use_cache: bool = True,
                lru_clear_cache: bool = False,
                lru_clear_arg_cache: bool = False,
                **kw: Any
            ) -> Any:
                key = KEY(args, kw)

                _clear_requested_cache_items(cache, lru_clear_cache, lru_clear_arg_cache, func, key)

                if not lru_use_cache:
                    logger.debug(
                        "Skipping cache for function %s with argument hash %s", func.__qualname__, key.__hash__
                    )
                    return await func(*args, **kw)

                if key in cache:
                    logger.debug(
                        "Using cached result for function %s with argument hash %s", func.__qualname__, key.__hash__
                    )
                    return cache[key]

                logger.debug(
                    "Executing function %s with argument hash %s and caching result", func.__qualname__, key.__hash__
                )
                cache[key] = await func(*args, **kw)
                return cache[key]

            return cast(F, wrapper_async)

        else:

            def wrapper_sync(
                *args: Any,
                lru_use_cache: bool = True,
                lru_clear_cache: bool = False,
                lru_clear_arg_cache: bool = False,
                **kw: Any
            ) -> Any:
                key = KEY(args, kw)

                _clear_requested_cache_items(cache, lru_clear_cache, lru_clear_arg_cache, func, key)

                if not lru_use_cache:
                    logger.debug(
                        "Skipping cache for function %s with argument hash %s", func.__qualname__, key.__hash__
                    )
                    return func(*args, **kw)

                if key in cache:
                    logger.debug(
                        "Using cached result for function %s with argument hash %s", func.__qualname__, key.__hash__
                    )
                    return cache[key]

                logger.debug(
                    "Executing function %s with argument hash %s and caching result", func.__qualname__, key.__hash__
                )
                cache[key] = func(*args, **kw)
                return cache[key]

            return cast(F, wrapper_sync)

    return decorator


def _clear_requested_cache_items(
    cache: LRU, lru_clear_cache: bool, lru_clear_arg_cache: bool, func: F, key: KEY
) -> None:
    if lru_clear_cache:
        cache.clear()
        logger.debug("Cleared cache for function %s", func.__qualname__)

    if lru_clear_arg_cache:
        cache.remove_by_key(key)
        logger.debug("Cleared argument cache for function %s with argument hash %s", func.__qualname__, key.__hash__)



from unittest import TestCase

# Tests another import - e.g call from multiple files in real life => should still be the same cache
from tests.data import reset_counter
from tests.data import sync_func
from tests.data import sync_func as sync_func_second_import
from tests.data import sync_func_clear_arg_cache, sync_func_clear_cache, sync_func_use_cache


class TestLruCache(TestCase):
    def test_lru_cache_with_cache_usage(self) -> None:
        # Arrange
        reset_counter()

        # Act
        result1 = sync_func_second_import("one", "one")
        result2 = sync_func_second_import("two", "two")
        result3 = sync_func_second_import("three", "three")

        result1_second = sync_func("one", "one")
        result2_second = sync_func("two", "two")
        result3_second = sync_func("three", "three")

        result4 = sync_func("four", "four")
        result_after_reached_max = sync_func("one", "one")

        # Assert
        self.assertEqual(result1, "one-one-1")
        self.assertEqual(result2, "two-two-2")
        self.assertEqual(result3, "three-three-3")

        self.assertEqual(result1_second, "one-one-1")
        self.assertEqual(result2_second, "two-two-2")
        self.assertEqual(result3_second, "three-three-3")

        self.assertEqual(result4, "four-four-4")
        # Now the counter has to be increased (indicating a method call) because one-one isn't cached anymore
        # due to the cache limit of 3 items
        self.assertEqual(result_after_reached_max, "one-one-5")

    def test_lru_cache_with_disabled_cache(self) -> None:
        # Arrange
        reset_counter()

        # Act
        result1 = sync_func_use_cache("one", "one", lru_use_cache=False)
        result2 = sync_func_use_cache("two", "two", lru_use_cache=False)
        result3 = sync_func_use_cache("three", "three", lru_use_cache=False)

        result1_second = sync_func_use_cache("one", "one", lru_use_cache=False)
        result2_second = sync_func_use_cache("two", "two", lru_use_cache=False)
        result3_second = sync_func_use_cache("three", "three", lru_use_cache=False)

        # Assert
        self.assertEqual(result1, "one-one-1")
        self.assertEqual(result2, "two-two-2")
        self.assertEqual(result3, "three-three-3")

        self.assertEqual(result1_second, "one-one-4")
        self.assertEqual(result2_second, "two-two-5")
        self.assertEqual(result3_second, "three-three-6")

    def test_lru_cache_with_clear_cache(self) -> None:
        # Arrange
        reset_counter()

        # Act
        result1 = sync_func_clear_cache("one", "one", lru_clear_cache=False)
        result2 = sync_func_clear_cache("two", "two", lru_clear_cache=False)

        result1_second = sync_func_clear_cache("one", "one", lru_clear_cache=False)
        result2_second = sync_func_clear_cache("two", "two", lru_clear_cache=False)

        result1_third = sync_func_clear_cache("one", "one", lru_clear_cache=True)  # Clear
        result2_third = sync_func_clear_cache("two", "two", lru_clear_cache=False)

        result1_fourth = sync_func_clear_cache("one", "one", lru_clear_cache=False)
        result2_fourth = sync_func_clear_cache("two", "two", lru_clear_cache=False)

        # Assert
        self.assertEqual(result1, "one-one-1")
        self.assertEqual(result2, "two-two-2")

        self.assertEqual(result1_second, "one-one-1")
        self.assertEqual(result2_second, "two-two-2")

        self.assertEqual(result1_third, "one-one-3")
        self.assertEqual(result2_third, "two-two-4")

        self.assertEqual(result1_fourth, "one-one-3")
        self.assertEqual(result2_fourth, "two-two-4")

    def test_lru_cache_with_clear_arg_cache(self) -> None:
        # Arrange
        reset_counter()

        # Act
        result1 = sync_func_clear_arg_cache("one", "one", lru_clear_arg_cache=True)  # May not fail, if not yet cached
        result2 = sync_func_clear_arg_cache("two", "two", lru_clear_arg_cache=False)

        result1_second = sync_func_clear_arg_cache("one", "one", lru_clear_arg_cache=False)
        result2_second = sync_func_clear_arg_cache("two", "two", lru_clear_arg_cache=False)

        result1_third = sync_func_clear_arg_cache("one", "one", lru_clear_arg_cache=True)  # Clear for args: one, one
        result2_third = sync_func_clear_arg_cache("two", "two", lru_clear_arg_cache=False)

        result1_fourth = sync_func_clear_arg_cache("one", "one", lru_clear_arg_cache=False)
        result2_fourth = sync_func_clear_arg_cache("two", "two", lru_clear_arg_cache=False)

        # Assert
        self.assertEqual(result1, "one-one-1")
        self.assertEqual(result2, "two-two-2")

        self.assertEqual(result1_second, "one-one-1")
        self.assertEqual(result2_second, "two-two-2")

        self.assertEqual(result1_third, "one-one-3")
        self.assertEqual(result2_third, "two-two-2")

        self.assertEqual(result1_fourth, "one-one-3")
        self.assertEqual(result2_fourth, "two-two-2")



from typing import Any, Union


class KEY:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.args = args
        self.kwargs = kwargs

        # Remove/ignore the decorator args
        kwargs.pop("use_cache", None)
        kwargs.pop("clear_cache", None)
        kwargs.pop("clear_arg_cache", None)

    def __eq__(self, obj: Any) -> bool:
        return hash(self) == hash(obj)

    def __hash__(self) -> int:
        def _hash(param: Any) -> Union[str, Any]:
            if isinstance(param, tuple):
                return tuple(map(_hash, param))
            if isinstance(param, dict):
                return tuple(map(_hash, param.items()))
            elif hasattr(param, "__dict__"):
                return str(vars(param))
            else:
                return str(param)

        return hash(_hash(self.args) + _hash(self.kwargs))



import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

with open("requirements.txt", "r") as fin:
    REQS = fin.read().splitlines()

with open("requirements-dev.txt", "r") as fin:
    REQS_DEV = [item for item in fin.read().splitlines() if not item.endswith(".txt")]

setuptools.setup(
    name="sag-py-cache-decorator",
    version="0.1.0",
    description="A decorator to cache method call results with similar parameters",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/SamhammerAG/sag_py_cache_decorator",
    author="Samhammer AG",
    author_email="support@samhammer.de",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development",
    ],
    keywords="cache, decorator",
    packages=setuptools.find_packages(exclude=["tests"]),
    package_data={"sag_py_cache_decorator": ["py.typed"]},
    python_requires=">=3.8",
    install_requires=REQS,
    extras_require={"dev": REQS_DEV},
    project_urls={
        "Documentation": "https://github.com/SamhammerAG/sag_py_cache_decorator",
        "Bug Reports": "https://github.com/SamhammerAG/sag_py_cache_decorator/issues",
        "Source": "https://github.com/SamhammerAG/sag_py_cache_decorator",
    },
)
